@@ -1,5 +1,4 @@
-{-# LANGUAGE OverloadedStrings, NoImplicitPrelude, FlexibleContexts, ScopedTypeVariables #-}
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE OverloadedStrings, FlexibleContexts, ScopedTypeVariables #-}
 module Network.HTTP.ReverseProxy
     ( -- * Types
       ProxyDest (..)
@@ -16,35 +15,51 @@ module Network.HTTP.ReverseProxy
     , wpsOnExc
     , wpsTimeout
     , wpsSetIpHeader
+    , wpsProcessBody
+    , wpsUpgradeToRaw
     , SetIpHeader (..)
+    {- FIXME
       -- * WAI to Raw
     , waiToRaw
+    -}
     ) where
 
-import ClassyPrelude
 import Data.Conduit
+import Data.Streaming.Network (readLens, AppData)
+import Data.Functor.Identity (Identity (..))
+import Data.Maybe (fromMaybe)
+import Control.Monad.Trans.Control (MonadBaseControl)
+import Data.Default.Class (def)
 import qualified Network.Wai as WAI
-import qualified Network.HTTP.Conduit as HC
-import Control.Exception.Lifted (try, finally)
-import Blaze.ByteString.Builder (fromByteString, flush)
-import Data.Word8 (isSpace, _colon, toLower, _cr)
+import qualified Network.HTTP.Client as HC
+import Network.HTTP.Client (BodyReader, brRead)
+import Control.Exception (bracket)
+import Blaze.ByteString.Builder (fromByteString)
+import Data.Word8 (isSpace, _colon, _cr)
+import qualified Data.ByteString as S
 import qualified Data.ByteString.Char8 as S8
 import qualified Network.HTTP.Types as HT
 import qualified Data.CaseInsensitive as CI
-import qualified Data.Text.Encoding as TE
 import qualified Data.Text.Lazy.Encoding as TLE
+import qualified Data.Text.Lazy as TL
 import qualified Data.Conduit.Network as DCN
 import Control.Concurrent.MVar.Lifted (newEmptyMVar, putMVar, takeMVar)
 import Control.Concurrent.Lifted (fork, killThread)
-import Control.Monad.Trans.Control (MonadBaseControl)
-import Network.Wai.Handler.Warp
-import Data.Conduit.Binary (sourceFileRange)
-import qualified Data.IORef as I
-import Network.Socket (PortNumber (PortNum), SockAddr (SockAddrInet))
-import Data.Default (Default (def))
-import Data.Version (showVersion)
-import qualified Paths_http_reverse_proxy
-import Network.Wai.Logger.Utils (showSockAddr)
+import Data.Default.Class (Default (..))
+import Network.Wai.Logger (showSockAddr)
+import qualified Data.Set as Set
+import Data.IORef
+import qualified Data.ByteString.Lazy as L
+import Control.Concurrent.Async (concurrently)
+import Blaze.ByteString.Builder (Builder, toLazyByteString)
+import Data.ByteString (ByteString)
+import Control.Monad.IO.Class (MonadIO, liftIO)
+import Control.Monad (unless, void)
+import Data.Monoid (mappend, (<>), mconcat)
+import Control.Exception.Lifted (try, SomeException, finally)
+import Control.Applicative ((<$>), (<|>))
+import Data.Set (Set)
+import qualified Data.Conduit.List as CL
 
 -- | Host\/port combination to which we want to proxy.
 data ProxyDest = ProxyDest
@@ -67,21 +82,28 @@ data ProxyDest = ProxyDest
 --
 -- If you need more control, such as modifying the request or response, use 'waiProxyTo'.
 rawProxyTo :: (MonadBaseControl IO m, MonadIO m)
-           => (HT.RequestHeaders -> m (Either (DCN.Application m) ProxyDest))
+           => (HT.RequestHeaders -> m (Either (DCN.AppData -> m ()) ProxyDest))
            -- ^ How to reverse proxy. A @Left@ result will run the given
            -- 'DCN.Application', whereas a @Right@ will reverse proxy to the
            -- given host\/port.
-           -> DCN.Application m
+           -> AppData -> m ()
 rawProxyTo getDest appdata = do
-    (rsrc, headers) <- fromClient $$+ getHeaders
+    (rsrc, headers) <- liftIO $ fromClient $$+ getHeaders
     edest <- getDest headers
     case edest of
         Left app -> do
             -- We know that the socket will be closed by the toClient side, so
             -- we can throw away the finalizer here.
-            (fromClient', _) <- unwrapResumable rsrc
-            app appdata { DCN.appSource = fromClient' }
-        Right (ProxyDest host port) -> DCN.runTCPClient (DCN.clientSettings port host) (withServer rsrc)
+            irsrc <- liftIO $ newIORef rsrc
+            let readData = do
+                    rsrc1 <- readIORef irsrc
+                    (rsrc2, mbs) <- rsrc1 $$++ await
+                    writeIORef irsrc rsrc2
+                    return $ fromMaybe "" mbs
+            app $ runIdentity (readLens (const (Identity readData)) appdata)
+
+
+        Right (ProxyDest host port) -> liftIO $ DCN.runTCPClient (DCN.clientSettings port host) (withServer rsrc)
   where
     fromClient = DCN.appSource appdata
     toClient = DCN.appSink appdata
@@ -98,10 +120,10 @@ rawProxyTo getDest appdata = do
 -- | Sends a simple 502 bad gateway error message with the contents of the
 -- exception.
 defaultOnExc :: SomeException -> WAI.Application
-defaultOnExc exc _ = return $ WAI.responseLBS
+defaultOnExc exc _ sendResponse = sendResponse $ WAI.responseLBS
     HT.status502
     [("content-type", "text/plain")]
-    ("Error connecting to gateway:\n\n" ++ TLE.encodeUtf8 (show exc))
+    ("Error connecting to gateway:\n\n" <> TLE.encodeUtf8 (TL.pack $ show exc))
 
 -- | The different responses that could be generated by a @waiProxyTo@ lookup
 -- function.
@@ -123,6 +145,10 @@ data WaiProxyResponse = WPRResponse WAI.Response
                         -- user.
                         --
                         -- Since 0.2.0
+                      | WPRApplication WAI.Application
+                        -- ^ Respond with the given WAI Application.
+                        --
+                        -- Since 0.4.0
 
 -- | Creates a WAI 'WAI.Application' which will handle reverse proxies.
 --
@@ -138,7 +164,7 @@ data WaiProxyResponse = WPRResponse WAI.Response
 -- Note: This function will use chunked request bodies for communicating with
 -- the proxied server. Not all servers necessarily support chunked request
 -- bodies, so please confirm that yours does (Warp, Snap, and Happstack, for example, do).
-waiProxyTo :: (WAI.Request -> ResourceT IO WaiProxyResponse)
+waiProxyTo :: (WAI.Request -> IO WaiProxyResponse)
            -- ^ How to reverse proxy. A @Left@ result will be sent verbatim as
            -- the response, whereas @Right@ will cause a reverse proxy.
            -> (SomeException -> WAI.Application)
@@ -157,6 +183,18 @@ data WaiProxySettings = WaiProxySettings
     -- Default: SIHFromSocket
     --
     -- Since 0.2.0
+    , wpsProcessBody :: HC.Response () -> Maybe (Conduit ByteString IO (Flush Builder))
+    -- ^ Post-process the response body returned from the host.
+    --
+    -- Since 0.2.1
+    , wpsUpgradeToRaw :: WAI.Request -> Bool
+    -- ^ Determine if the request should be upgraded to a raw proxy connection,
+    -- as is needed for WebSockets. Requires WAI 2.1 or higher and a WAI
+    -- handler with raw response support (e.g., Warp) to work.
+    --
+    -- Default: check if the upgrade header is websocket.
+    --
+    -- Since 0.3.1
     }
 
 -- | How to set the X-Real-IP request header.
@@ -171,69 +209,127 @@ instance Default WaiProxySettings where
         { wpsOnExc = defaultOnExc
         , wpsTimeout = Nothing
         , wpsSetIpHeader = SIHFromSocket
+        , wpsProcessBody = const Nothing
+        , wpsUpgradeToRaw = \req ->
+            (CI.mk <$> lookup "upgrade" (WAI.requestHeaders req)) == Just "websocket"
         }
 
-waiProxyToSettings getDest wps manager req0 = do
+renderHeaders :: WAI.Request -> HT.RequestHeaders -> Builder
+renderHeaders req headers
+    = fromByteString (WAI.requestMethod req)
+   <> fromByteString " "
+   <> fromByteString (WAI.rawPathInfo req)
+   <> fromByteString (WAI.rawQueryString req)
+   <> (if WAI.httpVersion req == HT.http11
+           then fromByteString " HTTP/1.1"
+           else fromByteString " HTTP/1.0")
+   <> mconcat (map goHeader headers)
+   <> fromByteString "\r\n\r\n"
+  where
+    goHeader (x, y)
+        = fromByteString "\r\n"
+       <> fromByteString (CI.original x)
+       <> fromByteString ": "
+       <> fromByteString y
+
+tryWebSockets :: WaiProxySettings -> ByteString -> Int -> WAI.Request -> (WAI.Response -> IO b) -> IO b -> IO b
+tryWebSockets wps host port req sendResponse fallback
+    | wpsUpgradeToRaw wps req =
+        sendResponse $ flip WAI.responseRaw backup $ \fromClientBody toClient ->
+            DCN.runTCPClient settings $ \server ->
+                let toServer = DCN.appSink server
+                    fromServer = DCN.appSource server
+                    fromClient = do
+                        mapM_ yield $ L.toChunks $ toLazyByteString headers
+                        let loop = do
+                                bs <- liftIO fromClientBody
+                                unless (S.null bs) $ do
+                                    yield bs
+                                    loop
+                        loop
+                    toClient' = awaitForever $ liftIO . toClient
+                    headers = renderHeaders req $ fixReqHeaders wps req
+                 in void $ concurrently
+                        (fromClient $$ toServer)
+                        (fromServer $$ toClient')
+    | otherwise = fallback
+  where
+    backup = WAI.responseLBS HT.status500 [("Content-Type", "text/plain")]
+        "http-reverse-proxy detected WebSockets request, but server does not support responseRaw"
+    settings = DCN.clientSettings port host
+
+strippedHeaders :: Set HT.HeaderName
+strippedHeaders = Set.fromList
+    ["content-length", "transfer-encoding", "accept-encoding", "content-encoding"]
+
+fixReqHeaders :: WaiProxySettings -> WAI.Request -> HT.RequestHeaders
+fixReqHeaders wps req =
+    addXRealIP $ filter (\(key, value) -> not $ key `Set.member` strippedHeaders
+                                       || (key == "connection" && value == "close"))
+               $ WAI.requestHeaders req
+  where
+    fromSocket = (("X-Real-IP", S8.pack $ showSockAddr $ WAI.remoteHost req):)
+    addXRealIP =
+        case wpsSetIpHeader wps of
+            SIHFromSocket -> fromSocket
+            SIHFromHeader ->
+                case lookup "x-real-ip" (WAI.requestHeaders req) <|> lookup "X-Forwarded-For" (WAI.requestHeaders req) of
+                    Nothing -> fromSocket
+                    Just ip -> (("X-Real-IP", ip):)
+            SIHNone -> id
+
+waiProxyToSettings :: (WAI.Request -> IO WaiProxyResponse)
+                   -> WaiProxySettings
+                   -> HC.Manager
+                   -> WAI.Application
+waiProxyToSettings getDest wps manager req0 sendResponse = do
     edest' <- getDest req0
     let edest =
             case edest' of
-                WPRResponse res -> Left res
+                WPRResponse res -> Left $ \_req -> ($ res)
                 WPRProxyDest pd -> Right (pd, req0)
                 WPRModifiedRequest req pd -> Right (pd, req)
+                WPRApplication app -> Left app
     case edest of
-        Left response -> return response
-        Right (ProxyDest host port, req) -> do
-            let req' = HC.def
+        Left app -> app req0 sendResponse
+        Right (ProxyDest host port, req) -> tryWebSockets wps host port req sendResponse $ do
+            let req' = def
                     { HC.method = WAI.requestMethod req
                     , HC.host = host
                     , HC.port = port
                     , HC.path = WAI.rawPathInfo req
                     , HC.queryString = WAI.rawQueryString req
-                    , HC.requestHeaders = filter (\(key, _) -> not $ key `member` strippedHeaders) $
-                        (case wpsSetIpHeader wps of
-                            SIHFromSocket -> (("X-Real-IP", S8.pack $ showSockAddr $ WAI.remoteHost req):)
-                            SIHFromHeader ->
-                                case lookup "x-real-ip" (WAI.requestHeaders req) <|> lookup "X-Forwarded-For" (WAI.requestHeaders req) of
-                                    Nothing -> id
-                                    Just ip -> (("X-Real-IP", ip):)
-                            SIHNone -> id)
-                        $ WAI.requestHeaders req
+                    , HC.requestHeaders = fixReqHeaders wps req
                     , HC.requestBody = body
                     , HC.redirectCount = 0
-#if MIN_VERSION_http_conduit(1, 9, 0)
                     , HC.checkStatus = \_ _ _ -> Nothing
-#else
-                    , HC.checkStatus = \_ _ -> Nothing
-#endif
                     , HC.responseTimeout = wpsTimeout wps
                     }
-                fbs bs = fromByteString bs <> flush
-                bodySrc = mapOutput fbs $ WAI.requestBody req
-                bodyChunked = HC.RequestBodySourceChunked bodySrc
-#if MIN_VERSION_wai(1, 4, 0)
                 body =
                     case WAI.requestBodyLength req of
-                        WAI.KnownLength i -> HC.RequestBodySource
+                        WAI.KnownLength i -> HC.RequestBodyStream
                             (fromIntegral i)
-                            bodySrc
-                        WAI.ChunkedBody -> bodyChunked
-#else
-                body = bodyChunked
-#endif
-            ex <- try $ HC.http req' manager
-            case ex of
-                Left e -> wpsOnExc wps e req
-                Right res -> do
-                    (src, _) <- unwrapResumable $ HC.responseBody res
-                    return $ WAI.ResponseSource
-                        (HC.responseStatus res)
-                        (filter (\(key, _) -> not $ key `member` strippedHeaders) $ HC.responseHeaders res) $ do
-                        yield Flush
-                        src =$= awaitForever (\bs -> yield (Chunk $ fromByteString bs) >> yield Flush)
-  where
-    strippedHeaders = asSet $ fromList ["content-length", "transfer-encoding", "accept-encoding", "content-encoding"]
-    asSet :: Set a -> Set a
-    asSet = id
+                            ($ WAI.requestBody req)
+                        WAI.ChunkedBody -> HC.RequestBodyStreamChunked ($ WAI.requestBody req)
+            bracket
+                (try $ HC.responseOpen req' manager)
+                (either (const $ return ()) HC.responseClose)
+                $ \ex -> do
+                case ex of
+                    Left e -> wpsOnExc wps e req sendResponse
+                    Right res -> do
+                        let conduit =
+                                case wpsProcessBody wps $ fmap (const ()) res of
+                                    Nothing -> awaitForever (\bs -> yield (Chunk $ fromByteString bs) >> yield Flush)
+                                    Just conduit' -> conduit'
+                            src = bodyReaderSource $ HC.responseBody res
+                        sendResponse $ WAI.responseStream
+                            (HC.responseStatus res)
+                            (filter (\(key, _) -> not $ key `Set.member` strippedHeaders) $ HC.responseHeaders res)
+                            (\sendChunk flush -> src $= conduit $$ CL.mapM_ (\mb ->
+                                case mb of
+                                    Flush -> flush
+                                    Chunk b -> sendChunk b))
 
 -- | Get the HTTP headers for the first request on the stream, returning on
 -- consumed bytes as leftovers. Has built-in limits on how many bytes it will
@@ -248,47 +344,39 @@ getHeaders =
       where
         close = leftover bs >> return bs
           where
-            bs = front empty
+            bs = front S8.empty
         push bs'
             | "\r\n\r\n" `S8.isInfixOf` bs
               || "\n\n" `S8.isInfixOf` bs
-              || length bs > 4096 = leftover bs >> return bs
-            | otherwise = go $ append bs
+              || S8.length bs > 4096 = leftover bs >> return bs
+            | otherwise = go $ mappend bs
           where
             bs = front bs'
-    toHeaders = map toHeader . takeWhile (not . null) . drop 1 . S8.lines
+    toHeaders = map toHeader . takeWhile (not . S8.null) . drop 1 . S8.lines
     toHeader bs =
         (CI.mk key, val)
       where
-        (key, bs') = break (== _colon) bs
-        val = takeWhile (/= _cr) $ dropWhile isSpace $ drop 1 bs'
+        (key, bs') = S.break (== _colon) bs
+        val = S.takeWhile (/= _cr) $ S.dropWhile isSpace $ S.drop 1 bs'
 
+{- FIXME
 -- | Convert a WAI application into a raw application, using Warp.
 waiToRaw :: WAI.Application -> DCN.Application IO
 waiToRaw app appdata0 =
-    loop $ transPipe lift fromClient0
+    loop fromClient0
   where
     fromClient0 = DCN.appSource appdata0
     toClient = DCN.appSink appdata0
     loop fromClient = do
-        mfromClient <- runResourceT $ do
-            ex <- try $ parseRequest conn 0 dummyAddr fromClient
+        mfromClient <- runResourceT $ withInternalState $ \internalState -> do
+            ex <- try $ parseRequest conn internalState dummyAddr fromClient
             case ex of
                 Left (_ :: SomeException) -> return Nothing
                 Right (req, fromClient') -> do
                     res <- app req
                     keepAlive <- sendResponse
-#if MIN_VERSION_warp(1, 3, 8)
                         defaultSettings
-                            { settingsServerName = S8.pack $ concat
-                                [ "Warp/"
-                                , warpVersion
-                                , " + http-reverse-proxy/"
-                                , showVersion Paths_http_reverse_proxy.version
-                                ]
-                            }
-#endif
-                        dummyCleaner req conn res
+                        req conn res
                     (fromClient'', _) <- liftIO fromClient' >>= unwrapResumable
                     return $ if keepAlive then Just fromClient'' else Nothing
         maybe (return ()) loop mfromClient
@@ -306,3 +394,14 @@ waiToRaw app appdata0 =
         , connClose = return ()
         , connRecv = error "connRecv should not be used"
         }
+        -}
+
+bodyReaderSource :: MonadIO m => BodyReader -> Source m ByteString
+bodyReaderSource br =
+    loop
+  where
+    loop = do
+        bs <- liftIO $ brRead br
+        unless (S.null bs) $ do
+            yield bs
+            loop
